#include <iostream>

#include "ace/Thread_Manager.h"
#include "orbsvcs/CosNamingC.h"
#include "orbsvcs/CosEventCommS.h"
#include "orbsvcs/CosEventChannelAdminS.h"

#include "EuroImpl.h"
#include "PushSupplierImpl.h"

using namespace std;

static const char *exception_message = "";
const char *servant_name = "Euro";
const char *push_supplier_name = "PushSupplier";
const char *naming_service_name = "NameService";  // Ne pas mettre d'autre nom que NameService
const char *event_service_name = "CosEventService";  // Ne pas mettre d'autre nom que CosEventService
const char *context_name = "EURO_context";

CosNaming::NamingContext_var root_naming_context = NULL;
CosEventChannelAdmin::ProxyPushConsumer_ptr pushConsumer = NULL;

/**
 * Permets d'enregistrer l'objet servant aupres du service de nommage
 * orb : ORB
 * name : nom de l'objet a enregistrer
 * obj : pointeur sur l'objet a enregistrer
 */
int bindObjectToNamingService(CORBA::ORB_ptr orb, const char *servant_name, CORBA::Object_ptr obj_servant)
{
	int error = 0;

	try
	{
		// Obtention d'une reference au service de nommage NamingService dont le nom est :
		// "NameService" (naming_service_name)
		exception_message = "Le service n'existe pas";
		CORBA::Object_var root_naming_context_obj = orb->resolve_initial_references(naming_service_name);

		// Obtention d'un contexte racine associe au service de nommage NamingService dont le nom est :
		// "NameService" (naming_service_name)
		exception_message = "Echec dans la restriction du contexte de nom";
		root_naming_context = CosNaming::NamingContext::_narrow(root_naming_context_obj);
		if (CORBA::is_nil(root_naming_context))
		{
			cerr  << exception_message << endl;
			return -1;
		}
	}
	catch(CORBA::ORB::InvalidName& ex)
	{
		// Ne devrait pas se produire !!
		ACE_ERROR ((LM_ERROR, "Server::bindObjectToNamingService - %s\n", exception_message));
		ACE_PRINT_EXCEPTION(ex, "Invalid Name Exception");
		return -1;
	}
	catch (CORBA::Exception& ex)
	{
		ACE_ERROR ((LM_ERROR, "Server::bindObjectToNamingService - %s\n", exception_message));
		ACE_PRINT_EXCEPTION(ex, "Exception");
		return -1;
	}

	try
	{
		// Liaison d'un contexte appele "EURO_context" (context_name) au contexte racine (root_naming_context)
		CosNaming::Name euro_naming_context_name;
		euro_naming_context_name.length(1);
		euro_naming_context_name[0].id = CORBA::string_dup(context_name);
		euro_naming_context_name[0].kind = CORBA::string_dup("Context");
		CosNaming::NamingContext_var euro_naming_context;

		try
		{
			euro_naming_context = root_naming_context->bind_new_context(euro_naming_context_name);
		}
		catch(CosNaming::NamingContext::AlreadyBound&)
		{
			// Si le contexte existe deja, cette exception sera lancee
			// Dans ce cas, il suffit de resoudre le nom de ce contexte et de recuperer le contexte
			// associe
			CORBA::Object_var obj = root_naming_context->resolve(euro_naming_context_name);
			euro_naming_context = CosNaming::NamingContext::_narrow(obj);
			if (CORBA::is_nil(euro_naming_context))
			{
				cerr << "Echec dans la restriction du contexte de nom lorsqu'il existe deja" << endl;
				return -1;
			}
		}

		// Liaison de notre objet obj_servant (le servant) au nom servant_name dans le contexte
		// euro_naming_context dont le nom est : "EURO_context" (context_name)
		CosNaming::Name obj_name;
		obj_name.length(1);
		obj_name[0].id = CORBA::string_dup(servant_name);
		obj_name[0].kind = CORBA::string_dup("Object");

		try
		{
			euro_naming_context->bind(obj_name, obj_servant);
		}
		catch(CosNaming::NamingContext::AlreadyBound&)
		{
			euro_naming_context->rebind(obj_name, obj_servant);
		}
	}
	catch (CORBA::Exception& ex)
	{
		ACE_ERROR ((LM_ERROR, "Server::bindObjectToNamingService - %s\n", exception_message));
		ACE_PRINT_EXCEPTION(ex, "Exception");
		return -1;
	}

	return error;
}

void *run_orb(void *arg)
{
	try
	{
		CORBA::ORB_ptr orb = ACE_reinterpret_cast(CORBA::ORB_ptr, arg);
		orb->run();
	}
	catch (CORBA::Exception& ex)
	{
		ACE_ERROR ((LM_ERROR, "Server::run_orb - %s\n", exception_message));
		ACE_PRINT_EXCEPTION(ex, "Exception");
		return (void *) -1;
	}

	return (void *) 0;
}

void deconnexion(int code = 0)
{
	cout << endl << "Deconnexion de l'objet : \"pushConsumer\"" << endl;
	pushConsumer->disconnect_push_consumer();

	cout << "Deconnexion du contexte : \"" << context_name << "\"" << endl;
	CosNaming::Name euro_naming_context_name;
	euro_naming_context_name.length(1);
	euro_naming_context_name[0].id = CORBA::string_dup(context_name);
	euro_naming_context_name[0].kind = CORBA::string_dup("Context");
	root_naming_context->unbind(euro_naming_context_name);

	exit(code);
}


int main(int argc, char *argv[])
{
	signal(SIGINT, deconnexion);

	try
	{
		// Etape 1 : Initialisation de l'ORB
		// ORB.init() permet de se connecter a l'ORB
		exception_message = "Etape 1 : Initialisation ORB";
		CORBA::ORB_var orb = CORBA::ORB_init(argc, argv, "TAO_ORB_1_4a");

		// Etape 2 : Obtention d'une reference sur le RootPOA (racine de l'arborescence
		// de POA)
		exception_message = "Etape 2 : Obtention reference sur RootPOA";
		CORBA::Object_var POA_object = orb->resolve_initial_references("RootPOA");
		PortableServer::POA_var root_POA = PortableServer::POA::_narrow(POA_object.in());

		// Etape 3 : Definition de la strategie qui s'applique aux differents objets geres par notre propre
		// POA que nous creons dans la foulee (par defaut : strategie TRANSIENT)
		// Remarque : la strategie PERSISTENT permet de garder la reference de l'objet meme si le serveur est arrete
		// puis redemarre.
		exception_message = "Etape 3 : Creation POA et strategie associee";
		CORBA::PolicyList policies;
		policies.length(1);
		policies[(CORBA::ULong)0] = root_POA->create_lifespan_policy(PortableServer::PERSISTENT);
		PortableServer::POA_var my_POA = root_POA->create_POA("EuroPOA", root_POA->the_POAManager(), policies);

		// Etape 4 : Creation du servant, c'est a dire de l'objet de type EuroImpl que nous souhaitons
		// rendre disponible a d'autres applications
		exception_message = "Etape 4 : Creation servant";
		EuroImpl servant;

		// Etape 5 : Activation du servant (notre objet de type EuroImpl dont le nom est "Euro" (servant_name))
		// dans notre POA (my_POA)
		// Il faut l'enregistrer et l'activer au sein du POA que nous venons de creer car ce n'est
		// pas un objet CORBA. C'est le POA (my_POA) qui va le publier et permettre aux applications
		// client d'y acceder.
		exception_message = "Etape 5 : Activation servant";
		PortableServer::ObjectId_var manager_id = PortableServer::string_to_ObjectId(servant_name);
		my_POA->activate_object_with_id(manager_id, &servant);

		// Etape 6 : Enregistrement de l'objet "Euro" (servant_name) qu'implemente la classe EuroImpl aupres du
		// service de nommage NamingService dont le nom est "NameService" (naming_service_name)
		// Les clients pourront alors se connecter a l'objet grace a se service de nommage
		int error = bindObjectToNamingService(orb, servant_name, servant._this());

		/***************************************************************************************************/
		/*********************************** Attachement a un canal d'evenements ***************************/
		/***************************************************************************************************/
		// Etape 7 : Connexion au service d'evenements
		exception_message = "Etape 7 : Connexion au service d'evenements";
		CORBA::Object_var root_naming_context_obj = orb->resolve_initial_references(naming_service_name);
		CosNaming::NamingContext_var root_naming_context
			= CosNaming::NamingContext::_narrow(root_naming_context_obj);

		exception_message = "Echec resolution du service d'evenements pour obtenir un canal d'evenements";
		CosNaming::Name event_service_obj_name;
		event_service_obj_name.length(1);
		event_service_obj_name[0].id = CORBA::string_dup(event_service_name);
		CORBA::Object_var event_service_obj = root_naming_context->resolve(event_service_obj_name);
		CosEventChannelAdmin::EventChannel_var eventChannel
			= CosEventChannelAdmin::EventChannel::_narrow(event_service_obj);

		cout << "Connexion au canal d'evenements effectuee" << endl;

		// Etape 8 : Creation de l'objet "PushSupplier"
		exception_message = "Etape 8 : Creation de l'objet \"PushSupplier\"";
		PushSupplierImpl pushSupplier(push_supplier_name, my_POA);

		// Etape 9 : Activation de notre objet "PushSupplier" (push_supplier_name) dans notre POA (my_POA)
		exception_message = "Etape 9 : Activation objet \"PushSupplier\"";
		PortableServer::ObjectId_var oid = PortableServer::string_to_ObjectId(push_supplier_name);
		my_POA->activate_object_with_id(oid, &pushSupplier);

		// Etape 10 : Activation du POA manager
		exception_message = "Etape 10 : Activation du POA manager";
		root_POA->the_POAManager()->activate();

		// Etape 11 : Connexion de l'objet "PushSupplier" au proxy
		exception_message = "Etape 11 : Connexion au proxy";
		pushConsumer = eventChannel->for_suppliers()->obtain_push_consumer();
		pushConsumer->connect_push_supplier(pushSupplier._this());
		cout << "PushSupplier connecte au proxy" << endl;

		// Etape 12 : Lancement de l'ORB dans la methode run_orb() threadee (asynchrone) afin que le serveur
		// ne soit pas bloque ici
		if (error == 0)
		{
			ACE_Thread_Manager::instance()->spawn((ACE_THR_FUNC)run_orb, (void *)orb, THR_DETACHED | THR_NEW_LWP);
			//ACE_Thread_Manager::instance()->wait();
			//orb->run();
		}
		else
		{
			return error;
		}

		/***************************************************************************************************/
		/***************************************** Envoi des evenements ************************************/
		/***************************************************************************************************/
		Convertisseur5::Event event_struct;
		event_struct.valeur = 1;
		while(true)
		{		
			CORBA::Any event;
			char message[100];

			strcpy(message, "");
			cout << "Entrer le message : ";
			cin >> message;

			// Un CTRL+C fait que message = ""
			// On met alors le thread courant en attente car un signal d'interruption (SIGINT)
			// va etre lance et il faut laisser la methode deconnexion() s'executer
			if (strcmp(message, "") == 0)
			{
				ACE_Thread_Manager::instance()->wait();
			}

			event_struct.message = CORBA::string_dup(message);

			// Noter l'utilisation geniale de l'operateur <<= qui permet de convertir n'importe
			// quel type vers le type CORBA::Any
			//event <<= message;  // Envoi d'une string uniquement
			event <<= event_struct;  // Envoi d'une structure
			pushConsumer->push(event);

			cout << "Evenement envoye au client" << endl;
			event_struct.valeur++;
		}

	}  // fin try
	catch (CORBA::UserException& userex)
	{
		ACE_ERROR ((LM_ERROR, "Server::main - %s\n", exception_message));
		ACE_PRINT_EXCEPTION(userex, "User Exception");
		return -1;
	}
	catch (CORBA::SystemException& sysex)
	{
		ACE_ERROR ((LM_ERROR, "Server::main - %s\n", exception_message));
		ACE_PRINT_EXCEPTION(sysex, "System Exception");
		return -1;
	}
	catch (CORBA::Exception& ex)
	{
		ACE_ERROR ((LM_ERROR, "Server::main - %s\n", exception_message));
		ACE_PRINT_EXCEPTION(ex, "Exception");
		return -1;
	}

	return 0;
}